/*
 * VGA display-related functions.
 *
 * Copyright (C) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __VGA_DISPLAY_H__
#define __VGA_DISPLAY_H__

#include "display.h"
#include "pic32_c.h"
#include "vga.h"



/* Display configuration type. */

typedef struct
{
    /* Current display state handler. */

    void (*state_handler)();

    /* Common transfer properties. */

    int line_timer;
    uint32_t output;

    /* DMA transfer properties. */

    int line_channels, transfer_int_num;
    uint32_t transfer_cell_size;

    /* Horizontal and vertical sync peripherals. */

    int hsync_unit, vsync_unit;

    /* Current scanline. */

    uint32_t line;

    /* Screen start address, frame limit, current pixel line. */

    uint8_t *screen_start, *screen_limit, *linedata;

    /* Screen size, used to wrap the current line pointer. */

    uint32_t screen_size;

    /* General display configuration. */

    display_config_t *display_config;

    /* Display line positions. */

    uint32_t hfreq_limit, hsync_start, hsync_end;

    /* Display region scanline positions. */

    uint32_t visible_start, vfp_start, vsync_start, vsync_end;

} vga_display_t;



/* Initialise a VGA display object with timing and pixel transfer details.
   This macro depends on the VGA timing constants being defined, and they are
   included above.

   VGA_Display(<name>)
*/

#define VGA_Display(NAME) \
    vga_display_t NAME = {                          \
        .transfer_cell_size = TRANSFER_CELL_SIZE,   \
        .hfreq_limit = HFREQ_LIMIT,                 \
        .hsync_start = HSYNC_START,                 \
        .hsync_end = HSYNC_END,                     \
        .visible_start = VISIBLE_START,             \
        .vfp_start = VFP_START,                     \
        .vsync_start = VSYNC_START,                 \
        .vsync_end = VSYNC_END};

/* Initialisation functions. */

void init_vga(display_config_t *display_config, int line_channels,
              int line_timer, int transfer_int_num);

void init_vga_with_timers(display_config_t *display_config, int line_channels,
                          int line_timer, int transfer_timer);

void vga_configure_sync(int hsync_unit, int vsync_unit);

void vga_configure_transfer(uint32_t output);

void vga_configure_dma_transfer(uint32_t output);

/* Initialisation helpers. */

void vga_configure_line_channel(int channel, int int_num, enum dma_chain chain,
                                uint32_t output);

void vga_configure_zero_channel(int channel, int int_num, int initiating,
                                uint32_t output);

/* Access-related operations. */

void vga_set_frame(display_config_t *display_config);
void vga_wait_visible(void);

/* Interrupt handlers. */

void vga_interrupt_handler(void);
void vga_hsync_interrupt_handler(void);
void vga_transfer_interrupt_handler(void);

/* Display state handlers. */

void vbp_active(void);
void visible_active(void);
void vfp_active(void);
void vsync_active(void);

/* DMA transfer operations. */

void start_visible(void);
void update_visible(void);
void stop_visible(void);
void update_transfers(int enable);

/* Vertical sync operations. */

void vsync_high(void);
void vsync_low(void);

#endif /* __VGA_DISPLAY_H__ */
